# Copyright (c) 2011 Adam Tauno Williams <awilliam@whitemice.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#
import hashlib
from sqlalchemy       import *
from coils.core import DocumentVersion, CoilsException
from coils.foundation import blob_manager_for_ds, BLOBManager

class BLOBCommand(object):

    def get_manager(self, document):
        manager = None
        if document.__entityName__ == 'Note':
            pass
            #TODO: provide a blob manager for Notes rather than the hard coded hack
        elif (document.project_id is not None):
            project = self._ctx.run_command('project::get', id = document.project_id)
            if (project is not None):
                # We are assuming this is a project document, so open it using a project path
                manager = blob_manager_for_ds(project.sky_url, project_id=project.object_id)
            else:
                pass
        else:
            # Document is not attached to a project, assuming this document is an attachment
            manager = blob_manager_for_ds('attachment:')
        return manager

    def get_handle(self, manager, mode, document, version=None, encoding='utf-8', create=False):
        path = manager.get_path(document, version=version)
        if (path is None):
            raise CoilsException('Unable to determine BLOB path for document revision objectId#{0} revision#{1}'.\
                    format(document.object_id, version))
        handle = BLOBManager.Open(path, mode, encoding=encoding, version=version, create=create)
        if (handle is None):
            raise CoilsException('Unable to open handle for document revision objectId#{0} revision#{1}'.\
                    format(document.object_id, version))
        return handle

    def write(self, in_handle, out_handle):
        in_handle.seek(0)
        m = hashlib.sha512()
        data = in_handle.read(4096)
        while data != '':
            m.update(data)
            out_handle.write(data)
            data = in_handle.read(4096)
        out_handle.flush()
        return (m.hexdigest(), in_handle.tell())

    def delete_version(self, manager, document, version=None):
        path = manager.get_path(document, version=version)
        BLOBManager.Delete(path)
        return True

    def set_context(self, document, folder=None, project=None, company=None, appointment=None):
        if (folder is not None):
            document.folder_id = folder.object_id
        if (project is not None):
            document.project_id = project.object_id
        if (company is not None):
            document.company_id = company.object_id
        if (appointment is not None):
            document.appointment_id = appointment.object_id

    def create_version(self, document, change_text=None, checksum=None):
        dv = DocumentVersion(document, change_text=change_text, checksum=checksum)
        #i = self._ctx.db_session().execute(Sequence('key_generator'))
        self._ctx.db_session().add(dv)
        self._ctx.db_session().flush()
        self._ctx.db_session().refresh(dv)
        self._ctx.db_session().refresh(document)
        return dv

    def store_to_version(self, manager, document, input_stream):
        version = self.create_version(document)
        # Copy the file into the target container (file)
        handle = self.get_handle(manager, 'w', document, version=document.version,
                                                         encoding='binary',
                                                         create=True)
        checksum, file_size = self.write(input_stream, handle)
        BLOBManager.Close(handle)
        # Set file size & checksum of Document entity
        document.set_file_size(file_size)
        document.set_checksum(checksum)
        # Set the file size & checksum of the Document Version entity
        version.set_checksum(checksum)
        version.set_file_size(file_size)

    def store_to_self(self, manager, document, input_stream):
        # TODO: This is not transactional; Hmmm, how to make it transactional?
        #       Perhaps we could create this as an aliased file name and move
        #       the data once the transaction compeletes [pending till commit?]
        #       Use the session_id as a filename suffix?
        handle = self.get_handle(manager, 'w', document, version=None,
                                                         encoding='binary',
                                                         create=True)
        checksum, file_size = self.write(input_stream, handle)
        BLOBManager.Close(handle)
        # Set file size & checksum of Document entity
        document.set_file_size(file_size)
        document.set_checksum(checksum)

